package nex

import (
	"bytes"
	"context"
	"errors"
	"io"
	"net"
	neturl "net/url"

	"tildegit.org/tjp/sliderule/internal/types"
)

// Client is used for sending nex requests and reading responses.
//
// It carries no state and is reusable simultaneously by multiple goroutines.
//
// The zero value is immediately usable.
type Client struct{}

// RoundTrip sends a single nex request and returns its response.
func (c Client) RoundTrip(ctx context.Context, request *types.Request) (*types.Response, error) {
	if request.Scheme != "nex" && request.Scheme != "" {
		return nil, errors.New("non-nex protocols not supported")
	}

	host := request.Host
	if _, port, _ := net.SplitHostPort(host); port == "" {
		host = net.JoinHostPort(host, "1900")
	}

	conn, err := (&net.Dialer{}).DialContext(ctx, "tcp", host)
	if err != nil {
		return nil, err
	}
	defer conn.Close()

	request.RemoteAddr = conn.RemoteAddr()
	request.TLSState = nil

	if _, err := conn.Write([]byte(request.Path + "\n")); err != nil {
		return nil, err
	}

	response, err := io.ReadAll(conn)
	if err != nil {
		return nil, err
	}

	return &types.Response{Body: bytes.NewBuffer(response), Request: request}, nil
}

// Fetch builds and sends a nex request, and returns the response.
func (c Client) Fetch(ctx context.Context, url string) (*types.Response, error) {
	u, err := neturl.Parse(url)
	if err != nil {
		return nil, err
	}
	return c.RoundTrip(ctx, &types.Request{URL: u})
}

func (c Client) IsRedirect(response *types.Response) bool { return false }
